﻿#region Usings
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.IO;

using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

using SeekThermal;
#endregion

namespace SeekThermal_Form
{
    public partial class MainForm : Form
    {
		#region <<<Vars>>>
        SeekThermalClass thermal;
        Thread thermalThread;
        int frameCount;
        bool stopThread;
        byte[] map_r = new byte[256];
        byte[] map_g = new byte[256];
      	byte[] map_b = new byte[256];
      	byte lastPalette = 0;
      	Bitmap Farbscala;
        ThermalFrame lastFrame;
        bool LockCtrl = false;
        delegate void Dele_V_Bitmap(Bitmap bmp);
        delegate void Dele_V_U16Array(ushort[] data);
        UInt16[] LastPreProcessed_Uint16;
		#endregion
		
		#region Form
        public MainForm()
        {
            InitializeComponent();
            Kernel_DrawPalette(); //init palette
        }
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            stopThread = true;
            if (thermal != null) {
            	if (thermal.Streaming) {
	        		thermal.Streaming=false; Thread.Sleep(500);
	        	}
                if (thermalThread!=null) {
        			thermalThread.Join(500);
        		}
                thermal.Deinit();
            }
        }
        
        void Btn_connectSeekClick(object sender, EventArgs e)
	    {
	        if (btn_connectSeek.BackColor==Color.LimeGreen) {
	        	stopThread = true;
	            if (thermal != null) {
		        	if (thermal.Streaming) {
		        		thermal.Streaming=false; Thread.Sleep(500);
		        	}
	        		if (thermalThread!=null) {
	        			thermalThread.Join(500);
	        		}
	                thermal.Deinit();
	            }
    			btn_GetProcImageStreaming.Text="Start Streaming";
    			btn_GetProcImageStreaming.BackColor=Color.Gainsboro;
    			btn_GetProcByteStreaming.Text="Start Streaming";
    			btn_GetProcByteStreaming.BackColor=Color.Gainsboro;
	        	btn_connectSeek.BackColor=Color.Gainsboro;
	        } else {
	        	btn_connectSeek.BackColor=Color.Gold; btn_connectSeek.Refresh();
	            var device = SeekThermalClass.Enumerate().First();
	            if(device == null) {
	            	btn_connectSeek.BackColor=Color.Red; btn_connectSeek.Refresh();
	                MessageBox.Show("No Seek Thermal devices found.");
	                return;
	            }
	            
	            thermal = new SeekThermalClass(device);
	            thermal.SeekStreamEvent += new SeekThermalClass.EventDelegate(Event_SeekStreamRecived);
	            
	            btn_connectSeek.BackColor=Color.LimeGreen;
	        }
            
        }
        void Event_SeekStreamRecived()
		{
        	BeginInvoke(new Dele_V_Bitmap(SeekStream_GetImage),new object[]{thermal.LastProcessedImage} );
        	BeginInvoke(new Dele_V_U16Array(SeekStream_GetByteArray),new object[]{thermal.lastProcessedFrame} );
		}
        #endregion
        
        #region Draw_Image/Palette
        void TrackBar_maxScroll(object sender, EventArgs e)
        {
        	if (LockCtrl) { return; }
        	if (trackBar_min.Value>trackBar_max.Value) {
        		trackBar_min.Value=trackBar_max.Value;
        	}
        	label_ScaleValues.Text="MaxValue: "+trackBar_max.Value.ToString()+
        		"\r\nMinValue: "+trackBar_min.Value.ToString();
        	Kernel_DrawImage();
        }
        void TrackBar_minScroll(object sender, EventArgs e)
        {
        	if (LockCtrl) { return; }
        	if (trackBar_min.Value>trackBar_max.Value) {
        		trackBar_max.Value=trackBar_min.Value;
        	}
        	label_ScaleValues.Text="MaxValue: "+trackBar_max.Value.ToString()+
        		"\r\nMinValue: "+trackBar_min.Value.ToString();
        	Kernel_DrawImage();
        }
        
        void Kernel_DrawImage()
        {
        	if (LastPreProcessed_Uint16==null) { return; }
        	bool Interpolate_x2 = chk_InterpolationX2.Checked;
        	PixelData Col = new PixelData();
        	UnsafeBitmap ubmp=null;
        	if (Interpolate_x2) {
        		ubmp = new UnsafeBitmap(206*2, 156*2);
        	} else {
        		ubmp = new UnsafeBitmap(206, 156);
        	}
        	ubmp.LockBitmap();
        	ushort max,min;
        	if (chk_AutoRange.Checked) {
                ushort val;
                min = 0xFFFF;
                max = 0x0000;
	            for (int y = 0; y < 156; y++) {
                    for (int x = 0; x < 206; x++) {
                        val = LastPreProcessed_Uint16[2+ y * 208 + x];
                        if (val < min) min = val;
                        if (val > max) max = val;
                    }
                }
        	} else {
        		max=(ushort)trackBar_max.Value;
        		min=(ushort)trackBar_min.Value;
        	}
        	int offset = (int)numericUpDown1.Value;
        	int center = ((max-min)/2)+min;
        	if (chk_useRefFrame.Checked) { //be shure, the reference is only active if valid
        		if (CalRefframe_Uint16==null) {
        			chk_useRefFrame.Checked=false;
        		}
        	}
            for (int y = 0; y < 156; y++) {
                for (int x = 0; x < 206; x++) {
                    int v = LastPreProcessed_Uint16[2+ y * 208 + x]; // + data.PixelData[y * 208 + 206] / 10; // no need to use column 207 since we already use frame id 4, max/min will be off if uncommented as well.
					
                    // Scale data to be within [0-255] for LUT mapping.
                    ushort maxmin = max;
                    if (maxmin>min) {
                    	maxmin -= min;
                    } else {
                    	maxmin=0;
                    }
                    
                    if (chk_useRefFrame.Checked) {
                    	v=LastPreProcessed_Uint16[2+ y * 208 + x]-CalRefframe_Uint16[2+ y * 208 + x]+min+offset;
                    }
                    
                    if (maxmin == 0) { maxmin = 1; }// Avoid divide by 0
                    v = (v - min) * 256 / maxmin;
                    if (chk_cal_useGainmap.Checked) {
                    	//v=(v+CalGainmap_Uint16[2+ y * 208 + x])/2;
                    	v=(int)(((double)(v)/(double)CalGainmap_processed[2+ y * 208 + x]));
                    }
                    if (v < 0) { v = 0; }
                    if (v > 255) {  v = 255; }

                    if (lastPalette>1) {
                    	Col.red=map_r[v];
                    	Col.green=map_g[v];
	                    Col.blue=map_b[v];
                    } else { // Greyscale output (would always be limited to 256 colors)
	                    Col.red=(byte)v;
	                    Col.green=(byte)v;
	                    Col.blue=(byte)v;
                    }
                    
                    if (Interpolate_x2) {
                    	ubmp.SetPixel(x * 2, y * 2, Col);
	                    ubmp.SetPixel(x * 2 + 1, y * 2, Col);
	                    ubmp.SetPixel(x * 2, y * 2 + 1, Col);
	                    ubmp.SetPixel(x * 2 + 1, y * 2 + 1, Col);
                    } else {
	                    ubmp.SetPixel(x, y, Col);
                    }
                }
            }
        	ubmp.UnlockBitmap();
			PicBoxMain.Image=ubmp.Bitmap;
        }
        //color palette
        void Rad_Pal_All_CheckedChanged(object sender, EventArgs e)
        {
        	Kernel_DrawPalette();
        }
        void panel_PalColAllClick(object sender, EventArgs e)
        {
        	if (colorDialog1.ShowDialog()==DialogResult.OK) {
        		Panel P = sender as Panel;
        		P.BackColor=colorDialog1.Color;
        		rad_Pal_SelfDual.Checked=true;
        		draw_dual_palette(panel_PalCol1.BackColor,panel_PalCol2.BackColor);
        	}
        }
        
        void Kernel_DrawPalette()
        {
        	byte pal = 0;
        	if (rad_Pal_Grayscale.Checked) { pal=1; }
        	if (rad_Pal_Ironbow.Checked) { pal=2; }
        	if (rad_Pal_Rainbow.Checked) { pal=3; }
        	if (rad_Pal_SelfDual.Checked) { pal=4; }
        	if (lastPalette==pal) { return; }
        	lastPalette=pal;
        	switch (lastPalette) {
        		case 1: draw_dual_palette(Color.White,Color.Black); break;
        		case 2: draw_IronBow_palette(); break;
        		case 3: draw_rainbow_palette(); break;
        		case 4: draw_dual_palette(panel_PalCol1.BackColor,panel_PalCol2.BackColor); break;
        	}
        	Kernel_DrawImage();
        }
        void draw_dual_palette(Color Hot_Col,Color Cold_Col)
		{
			//erstelle ein Bild mit 256 Pixeln, Farbwerte sind von 0-255 (byte)
			//also braucht man 1-256 Pixel zum auswerten
			Farbscala = new Bitmap(30, 256);
			//Grafikobjekte erstellen
			Graphics G = Graphics.FromImage(Farbscala);
			Rectangle rect = new Rectangle(0, 0, 30, 256);
			LinearGradientBrush GB = new LinearGradientBrush(rect, Hot_Col, Cold_Col, LinearGradientMode.Vertical);
			//fülle das rechteck mit den übergebenen farben
			G.FillRectangle(GB, rect);
			Bitmap img = (Bitmap)Farbscala;
			Color col = new Color();
			for (int i = 0; i < 256; i++ ) 
			{
				col = img.GetPixel(1,255-i);
				map_r[i] = col.R;
				map_g[i] = col.G;
				map_b[i] = col.B;
			}
			PicBoxPalette.Image=img;
		}
        void draw_IronBow_palette()
		{	//erstelle ein Bild mit 256 Pixeln, Farbwerte sind von 0-255 (byte)
			//also braucht man 1-256 Pixel zum auswerten
			Farbscala = new Bitmap(30, 256);
			//Grafikobjekte erstellen
			Graphics G = Graphics.FromImage(Farbscala);
			Rectangle rect = new Rectangle(0, 0, 30, 256);
			LinearGradientBrush GB = new LinearGradientBrush(rect, Color.Red, Color.Blue, LinearGradientMode.Vertical);
			//start und endfarbe mit neuen überschreiben
			ColorBlend CB = new ColorBlend();
			CB.Colors = new Color[]
			{
				Color.White,Color.Yellow,Color.Orange,Color.Crimson,Color.DarkViolet,Color.MediumBlue,Color.Black
			};
			//punkte festlegen, an denen die farben sein sollen, was dazwischen liegt
			//wird zum farbverlauf
      		float[] CP = new float[7];
      		//anfags und endpunkte müssen festliegen, 
      		//deshalb sind die anfangs und endfarben auch zweimal vorhanden
      		CP[0] = 0.0f;
      		CP[1] = 0.2f;
      		CP[2] = 0.3f;
      		CP[3] = 0.5f;
      		CP[4] = 0.7f;
      		CP[5] = 0.9f;
      		CP[6] = 1.0f;
      			
      		//werte übergeben
      		CB.Positions = CP;
			GB.InterpolationColors = CB;
			
			//fülle das rechteck mit den übergebenen farben
			G.FillRectangle(GB, rect);
			Bitmap img = (Bitmap)Farbscala;
			Color col = new Color();
			for (int i = 0; i < 256; i++ ) 
			{
				col = img.GetPixel(1,255-i);
				map_r[i] = col.R;
				map_g[i] = col.G;
				map_b[i] = col.B;
			}
			PicBoxPalette.Image=img;
		}
        void draw_rainbow_palette()
		{
			Farbscala = new Bitmap(30, 256);
			//Grafikobjekte erstellen
			Graphics G = Graphics.FromImage(Farbscala);
			Rectangle rect = new Rectangle(0, 0, 30, 256);
			LinearGradientBrush GB = new LinearGradientBrush(rect, Color.Red, Color.Blue, LinearGradientMode.Vertical);
			
			ColorBlend CB = new ColorBlend();
			CB.Colors = new Color[]
			{
				Color.White, Color.White,Color.Red,Color.Gold,Color.Yellow,
				Color.LimeGreen,Color.DodgerBlue,Color.DarkBlue,Color.Black,Color.Black
			};
      		float[] CP = new float[10];
      		CP[0] = 0.0f; CP[9] = 1.0f;
      		for (float i = 1; i < 9; i++ ) {	
      			CP[(int)i] = (float)( (i+-0.75)/7.5 );
      		}
      		CB.Positions = CP; GB.InterpolationColors = CB;
			
			//fülle das rechteck mit den übergebenen farben
			G.FillRectangle(GB, rect);
			Bitmap img = (Bitmap)Farbscala;
			Color col = new Color();
			for (int i = 0; i < 256; i++ ) 
			{
				col = img.GetPixel(1,255-i);
				map_r[i] = col.R;
				map_g[i] = col.G;
				map_b[i] = col.B;
			}
			PicBoxPalette.Image=img;
		}
		#endregion
		
        #region Get_ProcessedImage
        void Btn_GetFrame_asProcImageClick(object sender, EventArgs e)
        {
        	if (btn_connectSeek.BackColor!=Color.LimeGreen) { return; }
        	PicBoxMain.Image=thermal.GetFrame_asProcessedImage();
        }
        
        //Thread streaming is shared... was found in "Get_and_Process_ByteArray" -> Thread streaming
        void Btn_GetProcImageStreamingClick(object sender, EventArgs e)
        {
        	if (btn_connectSeek.BackColor!=Color.LimeGreen) { return; }
        	if (thermal.Streaming) {
        		thermal.Streaming=false;
        		btn_GetProcImageStreaming.Text="Start Streaming";
        		btn_GetProcImageStreaming.BackColor=Color.Gainsboro;
        	} else {
        		thermal.Start_ProcImage_Stream();
        		btn_GetProcImageStreaming.Text="Stop Streaming";
        		btn_GetProcImageStreaming.BackColor=Color.LimeGreen;
        	}
        }
        void SeekStream_GetImage(Bitmap bmp) 
        {
        	if (bmp==null) { return; }
        	if (btn_GetProcImageStreaming.BackColor==Color.LimeGreen) {
        		PicBoxMain.Image=bmp;
        	}
        }
        #endregion
        
        #region Get_and_Process_ByteArray
        UInt16[] GetCameraFrame()
        {
        	DateTime currentFrameTime = DateTime.Now;
            DateTime previousFrameTime = currentFrameTime;
            DateTime currentTime = DateTime.Now;
            DateTime previousTime = currentTime;
            
        	while (!stopThread && thermal != null) {
                // Store the previous frame time and compute the new one.
                previousFrameTime = currentFrameTime;
                currentFrameTime = DateTime.Now;
                // Get frame
                lastFrame = thermal.GetFrame_asClass();
                if (lastFrame.RawData==null) {
                	MessageBox.Show("No Frame Data recived."); return null;
                }
                // Time after frame capture
                previousTime = currentTime;
                currentTime = DateTime.Now;
                // Keep the first 6 frames, or anytime those frame IDs are encountered.
                // They might be usefull for image processing.
                switch (lastFrame.RawDataU16[10])
                {
                    case 4: thermal.FrameID4 = lastFrame; break;
//                    case 9: frameID9 = lastFrame; break;
//                    case 8: frameID8 = lastFrame; break;
//                    case 7: frameID7 = lastFrame; break;
//                    case 10: frameID10 = lastFrame; break;
//                    case 5: frameID5 = lastFrame; break;
                    default:
                        // Ignore the rest from safe keeping, since they are dealt with later.
                        // ID 1 is a calibration frame
                        // ID 6 is a pre-calibration frame
                        // ID 3 is a visual frame.
                        break;
                }
                
                if(lastFrame.IsCalibrationFrame) {
                    thermal.LastCalframe = lastFrame;
                } else {
                    if(thermal.LastCalframe != null && lastFrame.IsUsableFrame) {
                		frameCount++;
						return lastFrame.ProcessFrameU16(thermal.LastCalframe, thermal.FrameID4);
                    }
                }
            }
            return null;
        }
        
        void Kernel_PreProcess_Bytes(ushort[] data)
        {
        	int y;            
            int c = 0;
            ushort maxValue, minValue;

            // Display last sensor min/max values.
            label2.Text = "Max: " + data[0];
            label3.Text = "Min: " + data[1];

            label6.Text = "Max: " + thermal.LastCalframe.MaxValue;
            label7.Text = "Min: " + thermal.LastCalframe.MinValue;
			
            LockCtrl=true;
            if (chk_AutoRange.Checked) {
                maxValue = data[0];
                minValue = data[1];
                trackBar_max.Value = maxValue;
                trackBar_min.Value = minValue;
            } else {
                maxValue = (ushort)trackBar_max.Value;
                minValue = (ushort)trackBar_min.Value;
            }
            Application.DoEvents();
            LockCtrl=false;

            // Noise reduction
            //smaller number means less noise but also less details, 25 seems to be a good value
            // 25 was the original comment in frenky's sample code, since this is not the same code I think 4 is a good compromise.
            int noiseReduction = 4 * (data[0] - data[1]) / 256; 
            ushort[] arrColor = new ushort[5];
            ushort min, max;
            ulong temp,sum;
            ulong value;

            // Three passes works well, maybe too agressive.
        	for (y = 0; y < 154; y++) {
                for (int x = 0; x < 206; x++) {
                    min = 0xffff;
                    max = 0;
                    sum = 0;
                    //   2 
                    //0  d  1
                    //   3
                    arrColor[0] = data[2 + x + 208 * (y + 1)];
                    arrColor[1] = data[2 + x + 2 + 208 * (y + 1)];
                    arrColor[2] = data[2 + x + 1 + 208 * (y)];
                    arrColor[3] = data[2 + x + 1 + 208 * (y + 2)];
                    for (int i = 0; i < 4; i++) {
                        if (arrColor[i] > max) max = arrColor[i];
                        if (arrColor[i] < min) min = arrColor[i];
                        sum += arrColor[i];
                    }

                    //get average value, but exclude neighbour dead pixels from average:
                    temp = min;
                    temp += max;
                    value = sum;
                    value -= temp;
                    value /= 2;

                    arrColor[4] = data[2 + x + 1 + 208 * (y + 1)];
                    if (Math.Abs((ushort)value - arrColor[4]) > noiseReduction) {
                    	//if image is too different use average color from near pixels
                        data[2 + x + 1 + 208 * (y + 1)] = (ushort)value;
                    }
                }
            }

            // Clean up the borders so we don't have strange data
            // we loose the original value and we use the interpolation of the pixels next to the edge
            // for all the edges we care about (top bottom and left).
            ulong average;
            y = 0;
            for (int x = 1; x < 206; x++) {
                c = y * 208 + x;
                average = data[2+c + 207];
                average += data[2+c + 208];
                average += data[2+c + 209];
                average /= 3;
                data[2+c] = (ushort)average;
            }
            y = 155;
            for (int x = 1; x < 206; x++) {
                c = y * 208 + x;
                average = data[2+c - 207];
                average += data[2+c - 208];
                average += data[2+c - 209];
                average /= 3;
                data[c] = (ushort)average;
            }
            for (y = 1; y < 155; y++) {
                c = y * 208;
                average = data[2+c - 207];
                average += data[2+c + 1];
                average += data[2+c + 209];
                average /= 3;
                data[2+c] = (ushort)average;
            }
			
            // Recompute the min max after the image is all fixed
            // unless the sliders have overriden values, then it doesn't matter.
            if (chk_AutoRange.Checked) {
                ushort val;
                minValue = 0xFFFF;
                maxValue = 0x0000;
                for (y = 0; y < 156; y++)
                {
                    for (int x = 0; x < 206; x++)
                    {
                        val = data[2+ y * 208 + x];
                        if (val < minValue) minValue = val;
                        if (val > maxValue) maxValue = val;
                    }
                }
            }

            // Display displayed data min/max value.
            label9.Text = "Max: " + maxValue;
            label10.Text = "Min: " + minValue;
            
            //set to internal data array
            bool GrabWholeFrameInternal=true;
            if (chk_Average.Checked) {
	            if (LastPreProcessed_Uint16!=null) {
	            	if (LastPreProcessed_Uint16.Length==data.Length) {
	            		GrabWholeFrameInternal=false;
	            	}
	        	}
            }
            if (GrabWholeFrameInternal) {
            	LastPreProcessed_Uint16=data;
	            LastPreProcessed_Uint16[0]=maxValue;
	            LastPreProcessed_Uint16[1]=minValue;
            } else { //averraging
            	data[0]=maxValue;
	            data[1]=minValue;
            	for (int i = 0;i<data.Length ;i++ ) {
	            	long val = (long)LastPreProcessed_Uint16[i]*5L;
            		val+=data[i];
            		LastPreProcessed_Uint16[i]=(ushort)(val/6L);
            	} 
            }
        }
        
        void Btn_GetFrame_AsByteArrayClick(object sender, EventArgs e)
        {
        	stopThread=false;
        	UInt16[] data = GetCameraFrame();
            if (data == null) { return; }
            Kernel_PreProcess_Bytes(data);
            Kernel_DrawImage();
        }
        //Thread streaming
        void Btn_GetProcByteStreamingClick(object sender, EventArgs e)
        {
        	if (btn_connectSeek.BackColor!=Color.LimeGreen) { return; }
        	if (thermal.Streaming) {
        		thermal.Streaming=false;
        		btn_GetProcByteStreaming.Text="Start Streaming";
        		btn_GetProcByteStreaming.BackColor=Color.Gainsboro;
        	} else {
        		thermal.Start_ProcByte_Stream();
        		btn_GetProcByteStreaming.Text="Stop Streaming";
        		btn_GetProcByteStreaming.BackColor=Color.LimeGreen;
        	}
        }
        void SeekStream_GetByteArray(ushort[] data)
        {
        	if (data==null) { return; }
        	if (btn_GetProcByteStreaming.BackColor==Color.LimeGreen) {
        		Kernel_PreProcess_Bytes(data);
            	Kernel_DrawImage();
        	}
        }
        #endregion
        
        
        void Btn_GrabRAWFrameClick(object sender, EventArgs e)
        {
        	if (btn_connectSeek.BackColor!=Color.LimeGreen) { return; }
        	btn_GrabRAWFrame.BackColor=Color.Gold; btn_GrabRAWFrame.Refresh();
        	byte[] Frame = thermal.GetFrame_asBytes();
        	PixelData C = new PixelData();
        	UnsafeBitmap ubmp = new UnsafeBitmap(206,156);
        	ubmp.LockBitmap();
        	int min = Frame[1];//65525;
        	int max = Frame[0];//0;
//        	for (int y = 0; y < 154; y++) {
//                for (int x = 0; x < 206; x++) {
//        			int val = Frame[2+ y * 208 + x];
//        			if (val>max) { max=val; }
//        			if (val<min) { min=val; }
//                }
//            }
        	this.Text=min.ToString()+" to "+max.ToString();
        	for (int y = 0; y < 154; y++) {
                for (int x = 0; x < 206; x++) {
        			int val = Frame[2+ y * 208 + x];
        			int data = (int)(((float)(val-min)/(float)max)*255f);
        			if (data>255) { data=255; }
        			if (data<0) { data=0; }
        			C.red=(byte)data;
        			C.green=(byte)data;
        			C.blue=(byte)data;
        			ubmp.SetPixel(x,y,C);
                }
            }
        	ubmp.UnlockBitmap();
        	PicBoxMain.Image=ubmp.Bitmap;
        	btn_GrabRAWFrame.BackColor=Color.Gainsboro;
        }
        void Btn_SaveLastByteArrayToTxtClick(object sender, EventArgs e)
        {
        	if (LastPreProcessed_Uint16==null) { return; }
        	saveFileDialog1.FileName="LastPreProcessed_Uint16.txt";
        	if (saveFileDialog1.ShowDialog()!=DialogResult.OK) { return; }
        	
        	StreamWriter txt = new StreamWriter(saveFileDialog1.FileName);
        	for (int i =0;i<LastPreProcessed_Uint16.Length ;i++ ) {
        		txt.Write(LastPreProcessed_Uint16[i].ToString()+"#");
        	}
        	txt.Close();
        	
        	MessageBox.Show("File Saved:\r\n"+saveFileDialog1.FileName);
        }   
        
        #region Old_and_Test_Area
        
        void Trylater()
        {
        	Bitmap bm = new Bitmap(20,20);
        	
        	BitmapData srcData = bm.LockBits(
            new Rectangle(0, 0, bm.Width, bm.Height), 
            ImageLockMode.ReadOnly, 
            PixelFormat.Format32bppArgb);

			int stride = srcData.Stride;
			
			IntPtr Scan0 = srcData.Scan0;
			
			long[] totals = new long[] {0,0,0};
			
			int width = bm.Width;
			int height = bm.Height;
			
			unsafe
			{
			  byte* p = (byte*) (void*) Scan0;
			
			  for (int y = 0; y < height; y++)
			  {
			    for (int x = 0; x < width; x++)
			    {
			      for (int color = 0; color < 3; color++)
			      {
			        int idx = (y*stride) + x*4 + color;
			
			        totals[color] += p[idx];
			      }
			    }
			  }
			}
			
			int avgB = (int)totals[0] / (width*height);
			int avgG = (int)totals[1] / (width*height);
			int avgR = (int)totals[2] / (width*height);
        }
        
        void StartStream() 
        {
        	thermalThread = new Thread(ThermalThreadProc);
		    thermalThread.IsBackground = true;
		    thermalThread.Start();
        }
        void ThermalThreadProc()
        {
            DateTime currentFrameTime = DateTime.Now;
            DateTime previousFrameTime = currentFrameTime;
            DateTime currentTime = DateTime.Now;
            DateTime previousTime = currentTime;
//            int framesToCapture = 20;

            // Initialize frame (1 based)
            frameCount = 1;

            // Create the output files to save first 20 frames and associated metadata.
//            bw = new BinaryWriter(new FileStream("data.dat", FileMode.Create));
//            tw = new BinaryWriter(new FileStream("data.txt", FileMode.Create));

            while (!stopThread && thermal != null) {
                bool progress = false;

                // Store the previous frame time and compute the new one.
                previousFrameTime = currentFrameTime;
                currentFrameTime = DateTime.Now;

                // Get frame
                lastFrame = thermal.GetFrame_asClass();

                // Keep the first 6 frames, or anytime those frame IDs are encountered.
                // They might be usefull for image processing.
                switch (lastFrame.RawDataU16[10])
                {
                    case 4:
                        thermal.FrameID4 = lastFrame;
                        //frameID4.Inverse();
                        break;
//                    case 9:
//                        frameID9 = lastFrame;
//                        break;
//                    case 8:
//                        frameID8 = lastFrame;
//                        break;
//                    case 7:
//                        frameID7 = lastFrame;
//                        break;
//                    case 10:
//                        frameID10 = lastFrame;
//                        break;
//                    case 5:
//                        frameID5 = lastFrame;
//                        break;
                    default:
                        // Ignore the rest from safe keeping, since they are dealt with later.
                        //
                        // ID 1 is a calibration frame
                        // ID 6 is a pre-calibration frame
                        // ID 3 is a visual frame.
                        break;
                }

                // Time after frame capture
                previousTime = currentTime;
                currentTime = DateTime.Now;

                if(lastFrame.IsCalibrationFrame) {
                    thermal.LastCalframe = lastFrame;
                } else {
                    if(thermal.LastCalframe != null && lastFrame.IsUsableFrame) {
//						lastUsableFrame = lastFrame.ProcessFrameU16(thermal.LastCalframe, thermal.FrameID4);
                        progress = true;
                    }
                }

                // Toggle between automatic to manual range.
//                if (rangeOverrideToggle) {
//                    useAutomaticRange = !useAutomaticRange;
//                    rangeOverrideToggle = false;
//                }

                // Increase frame count.
                frameCount++;

                if(progress) {
                    Invalidate();
                }
            }
        }
        #endregion
        
        #region Calibration
        UInt16[] CalCold_Uint16;
        UInt16[] CalHot_Uint16;
        UInt16[] CalGainmap_Uint16;
        float[] CalGainmap_processed;
        bool[,] CalDefPixelMap=new bool[206,156];
        
        void Btn_SaveLastByteArrayToDatClick(object sender, EventArgs e)
        {
        	if (LastPreProcessed_Uint16==null) { return; }
        	saveFileDialog1.FileName="LastPreProcessed_Uint16.dat";
        	if (saveFileDialog1.ShowDialog()!=DialogResult.OK) { return; }
        	
        	FileStream FS = new FileStream(saveFileDialog1.FileName,FileMode.Create);
        	for (int i =0;i<LastPreProcessed_Uint16.Length ;i++ ) {
        		byte B1 = (byte)(LastPreProcessed_Uint16[i]>>8&0xff);
        		byte B2 = (byte)(LastPreProcessed_Uint16[i]&0xff);
        		FS.WriteByte(B1); FS.WriteByte(B2);
        	}
        	FS.Close();
        	
        	MessageBox.Show("File Saved:\r\n"+saveFileDialog1.FileName);
        }
        void Btn_LoadLastByteArrayFromDatClick(object sender, EventArgs e)
        {
        	if (openFileDialog1.ShowDialog()!=DialogResult.OK) { return; }
        	if (!File.Exists(openFileDialog1.FileName)) { return; }
        	FileStream FS = new FileStream(openFileDialog1.FileName,FileMode.Open);
        	byte[] inhalt = new byte[FS.Length];
        	FS.Read(inhalt,0,(int)FS.Length-1);
        	LastPreProcessed_Uint16=new ushort[FS.Length/2];
			FS.Close();
			
			int offset = 0;
        	for (int i =0;i<inhalt.Length-1 ;i+=2 ) {
				ushort val = (ushort)(inhalt[i]<<8|inhalt[i+1]);
				LastPreProcessed_Uint16[offset]=val; offset++;
        	}
        	
			tabControl1.SelectedIndex=0;
			Kernel_DrawImage();
        	//MessageBox.Show("File loadet:\r\n"+openFileDialog1.FileName);
        }
        
        void Btn_Cal_SetColdImageClick(object sender, EventArgs e)
        {
        	if (LastPreProcessed_Uint16==null) { MessageBox.Show("No Pre Processed Frame stored."); return; }
        	ushort[] data = new ushort[LastPreProcessed_Uint16.Length];
        	for (int i=0;i<data.Length ;i++ ) {
        		data[i]=LastPreProcessed_Uint16[i];
        	}
        	CalCold_Uint16=data;
        	btn_Cal_SetColdImage.BackColor=Color.LimeGreen;
        }
        void Btn_Cal_SetHotImageClick(object sender, EventArgs e)
        {
        	if (LastPreProcessed_Uint16==null) { MessageBox.Show("No Pre Processed Frame stored."); return; }
        	ushort[] data = new ushort[LastPreProcessed_Uint16.Length];
        	for (int i=0;i<data.Length ;i++ ) {
        		data[i]=LastPreProcessed_Uint16[i];
        	}
        	CalHot_Uint16=data;
        	btn_Cal_SetHotImage.BackColor=Color.LimeGreen;
        }
        
        void Kernel_Calibrate_Gainmap_DefPixMap()
        {
        	if (CalCold_Uint16==null) { MessageBox.Show("No Cold Frame stored."); return; }
        	if (CalHot_Uint16==null) { MessageBox.Show("No Hot Frame stored."); return; }
        	ushort minValue = 0xFFFF;
            ushort maxValue = 0x0000;
            int DefTresh = (int)num_Cal_DepPixTreshold.Value;
            CalGainmap_Uint16 = new ushort[CalCold_Uint16.Length];
        	for (int y = 0; y < 156; y++) {
                for (int x = 0; x < 206; x++) {
                    int VC = CalCold_Uint16[2+ y * 208 + x]; //Value Cold
					int VH = CalHot_Uint16[2+ y * 208 + x]; //Value Cold
					int VG = VH-VC;//value Gain
					if (VG>DefTresh) {
						CalDefPixelMap[x,y]=true; //is Defect Pixel
					} else {
						CalDefPixelMap[x,y]=false; //is Valid Pixel
					}
					CalGainmap_Uint16[2+ y * 208 + x]=(ushort)VG;
					if (VG < minValue) minValue = (ushort)VG;
                    if (VG > maxValue) maxValue = (ushort)VG;
                }
            }
            CalGainmap_Uint16[0]=maxValue;
            CalGainmap_Uint16[1]=minValue;
            
            //Save to File
            StreamWriter txt = new StreamWriter("Gainmap.txt");
        	for (int i =0;i<CalGainmap_Uint16.Length ;i++ ) {
        		txt.Write(CalGainmap_Uint16[i].ToString()+"#");
        	}
        	txt.Close();
        	
        	FileStream FS = new FileStream("Gainmap.dat",FileMode.Create);
        	for (int i =0;i<CalGainmap_Uint16.Length ;i++ ) {
        		byte B1 = (byte)(CalGainmap_Uint16[i]>>8&0xff);
        		byte B2 = (byte)(CalGainmap_Uint16[i]&0xff);
        		FS.WriteByte(B1); FS.WriteByte(B2);
        	}
        	FS.Close();
        	
        	StreamWriter txt2 = new StreamWriter("DefPixMap.map");
        	for (int y = 0; y < 156; y++) {
                for (int x = 0; x < 206; x++) {
        			if (CalDefPixelMap[x,y]) {
        				txt2.Write("#");
        			} else {
        				txt2.Write("-");
        			}
                }
        		txt2.WriteLine();
            }
        	txt2.Close();
        	
        	CalGainmap_processed=new float[CalGainmap_Uint16.Length];
			float max=CalGainmap_Uint16[0];
			float min=CalGainmap_Uint16[1];
			float HalfMax=(((CalGainmap_Uint16[0]-min)/max))/2;
			for (int i =0;i<CalGainmap_Uint16.Length ;i++ ) {
				float val = CalGainmap_Uint16[i];
				CalGainmap_processed[i]=(((val-min)/max))-HalfMax+1f;
			}
            
        	//Draw Image
        	Kernel_Draw_GainMap();
        	Kernel_Draw_DefPixMap();
        }
        void Kernel_Draw_GainMap()
        {
        	UnsafeBitmap ubmp = new UnsafeBitmap(206, 156);
        	PixelData Col = new PixelData();
        	ubmp.LockBitmap();
            for (int y = 0; y < 156; y++) {
                for (int x = 0; x < 206; x++) {
                    int v = CalGainmap_Uint16[2+ y * 208 + x]; // + data.PixelData[y * 208 + 206] / 10; // no need to use column 207 since we already use frame id 4, max/min will be off if uncommented as well.
					
                    // Scale data to be within [0-255] for LUT mapping.
                    ushort maxmin = CalGainmap_Uint16[0];
                    if (maxmin>CalGainmap_Uint16[1]) {
                    	maxmin -= CalGainmap_Uint16[1];
                    } else {
                    	maxmin=0;
                    }
                    if (maxmin == 0) { maxmin = 1; }// Avoid divide by 0
                    v = (v - CalGainmap_Uint16[1]) * 256 / maxmin;
                    if (v < 0) { v = 0; }
                    if (v > 255) {  v = 255; }
					
                    Col.red=(byte)v;
	                Col.green=(byte)v;
	                Col.blue=(byte)v;
                    
	                ubmp.SetPixel(x, y, Col);
                }
            }
        	ubmp.UnlockBitmap();
			PicBox_Gainmap.Image=ubmp.Bitmap;
        }
        void Kernel_Draw_DefPixMap()
        {
        	UnsafeBitmap ubmp = new UnsafeBitmap(206, 156);
        	PixelData Col = new PixelData();
        	ubmp.LockBitmap();
            for (int y = 0; y < 156; y++) {
                for (int x = 0; x < 206; x++) {
        			if (CalDefPixelMap[x,y]) {
	        			Col.red=0;
		                Col.green=0;
		                Col.blue=0;
        			} else {
        				Col.red=255;
		                Col.green=255;
		                Col.blue=255;
        			}
					
	                ubmp.SetPixel(x, y, Col);
                }
            }
        	ubmp.UnlockBitmap();
			PicBox_DefPixmap.Image=ubmp.Bitmap;
        }
        
        void Btn_CalibrateClick(object sender, EventArgs e)
        {
        	Kernel_Calibrate_Gainmap_DefPixMap();
        }
        
        //below the Mainimage
        UInt16[] CalRefframe_Uint16;
        void Btn_cal_RefframeGrabClick(object sender, EventArgs e)
        {
        	ushort[] data = new ushort[LastPreProcessed_Uint16.Length];
        	for (int i=0;i<data.Length ;i++ ) {
        		data[i]=LastPreProcessed_Uint16[i];
        	}
        	CalRefframe_Uint16=data;
        	chk_useRefFrame.Checked=true;
        }
        void Btn_cal_RefframeLoadClick(object sender, EventArgs e)
        {
        	if (!File.Exists("Cal_RefFrame.dat")) {
        		MessageBox.Show("no Reference stored (Cal_RefFrame.dat is missing)."); return;
        	}
        	FileStream FS = new FileStream("Cal_RefFrame.dat",FileMode.Open);
        	byte[] inhalt = new byte[FS.Length];
        	FS.Read(inhalt,0,(int)FS.Length-1);
        	CalRefframe_Uint16=new ushort[FS.Length/2];
			FS.Close();
			
			int offset = 0;
        	for (int i =0;i<inhalt.Length-1 ;i+=2 ) {
				ushort val = (ushort)(inhalt[i]<<8|inhalt[i+1]);
				CalRefframe_Uint16[offset]=val; offset++;
        	}
			chk_useRefFrame.Checked=true;
        }
        void Btn_cal_RefframeSaveClick(object sender, EventArgs e)
        {
        	if (LastPreProcessed_Uint16==null) {
        		MessageBox.Show("no data to save.");
        	}
        	FileStream FS = new FileStream("Cal_RefFrame.dat",FileMode.Create);
        	for (int i =0;i<LastPreProcessed_Uint16.Length ;i++ ) {
        		byte B1 = (byte)(LastPreProcessed_Uint16[i]>>8&0xff);
        		byte B2 = (byte)(LastPreProcessed_Uint16[i]&0xff);
        		FS.WriteByte(B1); FS.WriteByte(B2);
        	}
        	FS.Close();
        	
        	MessageBox.Show("File Saved:\r\nCal_RefFrame.dat");
        }
        #endregion
        
        
        void Btn_cal_loadMapsClick(object sender, EventArgs e)
        {
        	FileStream FS = new FileStream("Gainmap.dat",FileMode.Open);
        	byte[] inhalt = new byte[FS.Length];
        	FS.Read(inhalt,0,(int)FS.Length-1);
        	CalGainmap_Uint16=new ushort[FS.Length/2];
			FS.Close();
			
			int offset = 0;
        	for (int i =0;i<inhalt.Length-1 ;i+=2 ) {
				ushort val = (ushort)(inhalt[i]<<8|inhalt[i+1]);
				CalGainmap_Uint16[offset]=val; offset++;
        	}
			CalGainmap_processed=new float[CalGainmap_Uint16.Length];
			float max=CalGainmap_Uint16[0];
			float min=CalGainmap_Uint16[1];
			float HalfMax=(((CalGainmap_Uint16[0]-min)/max))/2;
			for (int i =0;i<CalGainmap_Uint16.Length ;i++ ) {
				float val = CalGainmap_Uint16[i];
				CalGainmap_processed[i]=(((val-min)/max))-HalfMax+1f;
			}
			Kernel_Draw_GainMap();
        }
        void NumericUpDown1ValueChanged(object sender, EventArgs e)
        {
        	Kernel_DrawImage();
        }
        
        
        
        
        
        
        
    }
}
